from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import MagicMock

import yaml
from ops.charm import CharmBase, CharmEvents
from ops.framework import EventBase

from scenario.runtime import Runtime
from scenario.state import Event, State, _CharmSpec


def charm_type():
    class _CharmEvents(CharmEvents):
        pass

    class MyCharm(CharmBase):
        on = _CharmEvents()
        _event = None

        def __init__(self, framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._catchall)

        def _catchall(self, e):
            MyCharm._event = e

    return MyCharm


def test_event_hooks():
    with TemporaryDirectory() as tempdir:
        meta = {
            "name": "foo",
            "requires": {"ingress-per-unit": {"interface": "ingress_per_unit"}},
        }
        temppath = Path(tempdir)
        meta_file = temppath / "metadata.yaml"
        meta_file.write_text(yaml.safe_dump(meta))

        runtime = Runtime(
            _CharmSpec(
                charm_type(),
                meta=meta,
            )
        )

        pre_event = MagicMock(return_value=None)
        post_event = MagicMock(return_value=None)
        runtime.exec(
            state=State(),
            event=Event("update_status"),
            pre_event=pre_event,
            post_event=post_event,
        )

        assert pre_event.called
        assert post_event.called


def test_event_emission():
    with TemporaryDirectory() as tempdir:
        meta = {
            "name": "foo",
            "requires": {"ingress-per-unit": {"interface": "ingress_per_unit"}},
        }

        my_charm_type = charm_type()

        class MyEvt(EventBase):
            pass

        my_charm_type.on.define_event("bar", MyEvt)

        runtime = Runtime(
            _CharmSpec(
                my_charm_type,
                meta=meta,
            ),
        )

        runtime.exec(state=State(), event=Event("bar"))

        assert my_charm_type._event
        assert isinstance(my_charm_type._event, MyEvt)



import typing
from itertools import chain
from typing import Any, Callable, Dict, Iterable, Optional, TextIO, Type, Union

from scenario.logger import logger as scenario_logger
from scenario.state import (
    ATTACH_ALL_STORAGES,
    BREAK_ALL_RELATIONS,
    CREATE_ALL_RELATIONS,
    DETACH_ALL_STORAGES,
    META_EVENTS,
    Event,
    InjectRelation,
    State,
)

if typing.TYPE_CHECKING:
    from ops.testing import CharmType

CharmMeta = Optional[Union[str, TextIO, dict]]

logger = scenario_logger.getChild("scenario")


def decompose_meta_event(meta_event: Event, state: State):
    # decompose the meta event

    if meta_event.name in [ATTACH_ALL_STORAGES, DETACH_ALL_STORAGES]:
        logger.warning(f"meta-event {meta_event.name} not supported yet")
        return

    if meta_event.name in [CREATE_ALL_RELATIONS, BREAK_ALL_RELATIONS]:
        for relation in state.relations:
            event = Event(
                relation.endpoint + META_EVENTS[meta_event.name],
                args=(
                    # right now, the Relation object hasn't been created by ops yet, so we can't pass it down.
                    # this will be replaced by a Relation instance before the event is fired.
                    InjectRelation(relation.endpoint, relation.relation_id),
                ),
            )
            logger.debug(f"decomposed meta {meta_event.name}: {event}")
            yield event, state.copy()

    else:
        raise RuntimeError(f"unknown meta-event {meta_event.name}")


def generate_startup_sequence(state_template: State):
    yield from chain(
        decompose_meta_event(Event(ATTACH_ALL_STORAGES), state_template.copy()),
        ((Event("start"), state_template.copy()),),
        decompose_meta_event(Event(CREATE_ALL_RELATIONS), state_template.copy()),
        (
            (
                Event(
                    "leader_elected"
                    if state_template.leader
                    else "leader_settings_changed"
                ),
                state_template.copy(),
            ),
            (Event("config_changed"), state_template.copy()),
            (Event("install"), state_template.copy()),
        ),
    )


def generate_teardown_sequence(state_template: State):
    yield from chain(
        decompose_meta_event(Event(BREAK_ALL_RELATIONS), state_template.copy()),
        decompose_meta_event(Event(DETACH_ALL_STORAGES), state_template.copy()),
        (
            (Event("stop"), state_template.copy()),
            (Event("remove"), state_template.copy()),
        ),
    )


def generate_builtin_sequences(template_states: Iterable[State]):
    for template_state in template_states:
        yield from chain(
            generate_startup_sequence(template_state),
            generate_teardown_sequence(template_state),
        )


def check_builtin_sequences(
    charm_type: Type["CharmType"],
    meta: Optional[Dict[str, Any]] = None,
    actions: Optional[Dict[str, Any]] = None,
    config: Optional[Dict[str, Any]] = None,
    template_state: State = None,
    pre_event: Optional[Callable[["CharmType"], None]] = None,
    post_event: Optional[Callable[["CharmType"], None]] = None,
):
    """Test that all the builtin startup and teardown events can fire without errors.

    This will play both scenarios with and without leadership, and raise any exceptions.

    This is a baseline check that in principle all charms (except specific use-cases perhaps),
    should pass out of the box.

    If you want to, you can inject more stringent state checks using the
    pre_event and post_event hooks.
    """

    template = template_state if template_state else State()

    for event, state in generate_builtin_sequences(
        (
            template.replace(leader=True),
            template.replace(leader=False),
        )
    ):
        state.trigger(
            event=event,
            charm_type=charm_type,
            meta=meta,
            actions=actions,
            config=config,
            pre_event=pre_event,
            post_event=post_event,
        )



### This file contains stuff that ideally should be in ops.
# see https://github.com/canonical/operator/pull/862

import inspect
import os
from typing import TYPE_CHECKING, Callable, Optional

import ops.charm
import ops.framework
import ops.model
import ops.storage
from ops.charm import CharmMeta
from ops.log import setup_root_logging
from ops.main import CHARM_STATE_FILE, _Dispatcher, _emit_charm_event, _get_charm_dir

from scenario.logger import logger as scenario_logger
from scenario.mocking import _MockModelBackend

if TYPE_CHECKING:
    from ops.testing import CharmType

    from scenario.state import Event, State, _CharmSpec

logger = scenario_logger.getChild("ops_main_mock")


class NoObserverError(RuntimeError):
    """Error raised when the event being dispatched has no registered observers."""


def main(
    pre_event: Optional[Callable[["CharmType"], None]] = None,
    post_event: Optional[Callable[["CharmType"], None]] = None,
    state: "State" = None,
    event: "Event" = None,
    charm_spec: "_CharmSpec" = None,
):
    """Set up the charm and dispatch the observed event."""
    charm_class = charm_spec.charm_type
    charm_dir = _get_charm_dir()
    model_backend = _MockModelBackend(  # pyright: reportPrivateUsage=false
        state=state, event=event, charm_spec=charm_spec
    )
    debug = "JUJU_DEBUG" in os.environ
    setup_root_logging(model_backend, debug=debug)
    logger.debug(
        "Operator Framework %s up and running.", ops.__version__
    )  # type:ignore

    dispatcher = _Dispatcher(charm_dir)
    dispatcher.run_any_legacy_hook()

    metadata = (charm_dir / "metadata.yaml").read_text()
    actions_meta = charm_dir / "actions.yaml"
    if actions_meta.exists():
        actions_metadata = actions_meta.read_text()
    else:
        actions_metadata = None

    meta = CharmMeta.from_yaml(metadata, actions_metadata)
    model = ops.model.Model(meta, model_backend)

    charm_state_path = charm_dir / CHARM_STATE_FILE

    # TODO: add use_juju_for_storage support
    store = ops.storage.SQLiteStorage(charm_state_path)
    framework = ops.framework.Framework(store, charm_dir, meta, model)
    framework.set_breakpointhook()
    try:
        sig = inspect.signature(charm_class)
        sig.bind(framework)  # signature check

        charm = charm_class(framework)
        dispatcher.ensure_event_links(charm)

        # Skip reemission of deferred events for collect-metrics events because
        # they do not have the full access to all hook tools.
        if not dispatcher.is_restricted_context():
            framework.reemit()

        if pre_event:
            pre_event(charm)

        if not getattr(charm.on, dispatcher.event_name, None):
            raise NoObserverError(
                f"Charm has no registered observers for {dispatcher.event_name!r}. "
                f"This is probably not what you were looking for."
            )

        _emit_charm_event(charm, dispatcher.event_name)

        if post_event:
            post_event(charm)

        framework.commit()
    finally:
        framework.close()
